
HARD_KEYWORDS = (
    "as", "break", "catch", "class", "continue", "do", "else", "false", "for", "fun", "if", "in", "interface", "is",
    "null", "object", "package", "return", "super", "this", "throw", "true", "try", "typealias", "typeof",
    "val", "var", "when", "while")

SOFT_KEYWORDS = (
    "by", "constructor", "delegate", "dynamic", "field", "file", "finally", "get", "import", "init",
    "it", "param", "property", "receiver", "set", "setparam", "where")

MODIFIER_KEYWORDS = (
    "actual", "abstract", "annotation", "companion", "const", "crossinline", "data", "enum", "expect",
    "external", "final", "infix", "inline", "inner", "internal", "lateinit", "noinline", "open", "operator",
    "out", "override", "private", "protected", "public", "reified", "sealed", "suspend", "tailrec", "vararg")
